/*
 * Simple verification tool for fcache.
 *
 * gcc -Wall -O2 -g -D_GNU_SOURCE -D_LARGEFILE_SOURCE -D_FILE_OFFSET_BITS=64 -o fcache_fsck fcache_fsck.c
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <byteswap.h>
#include <endian.h>

#include "list.h"

#if __WORDSIZE == 32
typedef unsigned long long u64;
#else
typedef unsigned long u64;
#endif

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define be32_to_cpu(x)		__bswap_32(x)
#define be64_to_cpu(x)		__bswap_64(x)
#else
#define be32_to_cpu(x)		(x)
#define be64_to_cpu(x)		(x)
#endif

#define FCACHE_MAGIC    0x61786663
#define FCACHE_VERSION  0x03

typedef unsigned short u16;
typedef unsigned int u32;

/*
 * Maps a file system block to the fcache
 */
struct fcache_extent {
	u64 fs_sector;		/* real device offset */
	u64 cache_sector;	/* cache device offset */
	u32 fs_size;		/* extent length in bytes */
};

#define BDEVNAME_SIZE	32

/*
 * Header on fcache device
 */
struct fcache_header {
	u32 magic;              /* bdcache magic */
	u32 version;            /* bdcache version */
	u32 nr_extents;         /* nr of extents in cache */
	u32 max_extents;        /* max nr of extents */
	u32 serial;             /* fs and cache serial */
	u32 extent_offset;      /* where extents start */
	u64 fs_start_sector;
	u64 fs_sectors;
	char fs_dev[BDEVNAME_SIZE];
	u64 cache_blocks;	/* total cache blocks */
	u64 cache_blocks_used;	/* used cache blocks */
};

struct fcache_li {
	struct fcache_extent fe;
	struct list_head fs_list;
	struct list_head cache_list;
};

#define PAGE_SIZE	4096
#define PAGE_MASK	(PAGE_SIZE - 1)

static int good_extents, bad_extents;
static int fsfd, fd;
static LIST_HEAD(fe_fs_list);
static LIST_HEAD(fe_cache_list);

#define ALIGN(buf)      (void *) (((unsigned long) (buf) + PAGE_MASK) & ~(PAGE_MASK))

static inline int error(const char *e)
{
	perror(e);
	return -1;
}

static void sort_fs_list_item(struct fcache_li *li)
{
	struct list_head *list = &fe_fs_list;
	struct list_head *node;
	struct fcache_li *__li;

	node = list;
	while ((node = node->prev) != list) {
		__li = list_entry(node, struct fcache_li, fs_list);

		if (__li->fe.fs_sector < li->fe.fs_sector)
			break;
	}

	INIT_LIST_HEAD(&li->fs_list);
	list_add(&li->fs_list, node);
}

static void sort_cache_list_item(struct fcache_li *li)
{
	struct list_head *list = &fe_cache_list;
	struct list_head *node;
	struct fcache_li *__li;

	node = list;
	while ((node = node->prev) != list) {
		__li = list_entry(node, struct fcache_li, cache_list);

		if (__li->fe.cache_sector < li->fe.cache_sector)
			break;
	}

	INIT_LIST_HEAD(&li->cache_list);
	list_add(&li->cache_list, node);
}

static void add_list_item(struct fcache_extent *fe)
{
	struct fcache_li *li = malloc(sizeof(*li));

	li->fe = *fe;

	sort_fs_list_item(li);
	sort_cache_list_item(li);
}

static void hexdump(unsigned char *buffer, int len, char *msg)
{
	int i;

	printf("\t%s: ", msg);
	for (i = 0; i < len; i++)
		printf("%02x ", buffer[i]);
	printf("\n");
}

static int check_extent_data(struct fcache_extent *fe)
{
	off_t cur_fd_pos = lseek(fd, 0, SEEK_CUR);
	unsigned char *buffer, *fsbuffer;
	int ret, i;
	off_t pos;

	if (fsfd == 0)
		return 0;

	if (cur_fd_pos < 0)
		return error("lseek");

	buffer = ALIGN(malloc(fe->fs_size + PAGE_SIZE - 1));
	fsbuffer = ALIGN(malloc(fe->fs_size + PAGE_SIZE - 1));

	pos = ((unsigned long long) fe->cache_sector) << 9;
	if (lseek(fd, pos, SEEK_SET) < 0)
		return error("lseek");
	ret = read(fd, buffer, fe->fs_size);
	if (ret != fe->fs_size)
		return error("read");

	pos = ((unsigned long long) fe->fs_sector) << 9;
	if (lseek(fsfd, pos, SEEK_SET) < 0)
		return error("lseek");
	ret = read(fsfd, fsbuffer, fe->fs_size);
	if (ret != fe->fs_size)
		return error("read");

	if (lseek(fd, cur_fd_pos, SEEK_SET) < 0)
		return error("lseek");

	if (!memcmp(buffer, fsbuffer, fe->fs_size))
		return 0;

	/*
	 * Uh oh, content differs. Check where we run into problems.
	 */
	for (i = 0; i < fe->fs_size; i++) {
		if (buffer[i] == fsbuffer[i])
			continue;

		printf(": differ at offset %u\n", i);
		hexdump(buffer, 20, "cache");
		hexdump(fsbuffer, 20, "fs   ");
		break;
	}

	return 1;
}

static void __show_extents(void *buf, int max_extents)
{
	struct fcache_extent *fe;
	int i;

	for (i = 0; i < max_extents; i++) {
		fe = buf + i * sizeof(*fe);

		fe->fs_sector = be64_to_cpu(fe->fs_sector);
		fe->cache_sector = be64_to_cpu(fe->cache_sector);
		fe->fs_size = be32_to_cpu(fe->fs_size);
	
		if (!fe->fs_size || (fe->fs_size & 511)) {
			printf("\t%d looks bad\n", i);
			break;
		}

		printf("\t%8llu -> %8llu (%4u sectors)", (unsigned long long) fe->fs_sector, (unsigned long long) fe->cache_sector, fe->fs_size >> 9);

		if (check_extent_data(fe))
			bad_extents++;
		else {
			if (fsfd)
				printf(": content verified OK\n");
			else
				printf("\n");
			good_extents++;
		}
		add_list_item(fe);
	}
}

static int show_extents(struct fcache_header *hdr)
{
	char *buffer = ALIGN(malloc(2 * PAGE_SIZE - 1));
	off_t extent_off;

	printf("\nExtents:\n");
	extent_off = PAGE_SIZE;
	do {
		int nr_extents = PAGE_SIZE / sizeof(struct fcache_extent);

		if (nr_extents > hdr->nr_extents)
			nr_extents = hdr->nr_extents;

		memset(buffer, 0, PAGE_SIZE);

		if (lseek(fd, extent_off, SEEK_SET) < 0)
			return error("lseek");

		if (read(fd, buffer, PAGE_SIZE) != PAGE_SIZE)
			return error("read");

		__show_extents(buffer, nr_extents);
		hdr->nr_extents -= nr_extents;
		extent_off += PAGE_SIZE;
	} while (hdr->nr_extents);

	return 0;
}

static inline int mergeable(struct fcache_li *prv, struct fcache_li *nxt)
{
	if ((prv->fe.fs_sector + (prv->fe.fs_size >> 9)) != nxt->fe.fs_sector)
		return 0;
	if ((prv->fe.cache_sector + (prv->fe.fs_size >> 9)) != nxt->fe.cache_sector)
		return 0;

	return 1;
}

static int verify_layout(int *merges)
{
	struct fcache_li *li, *last_li;
	struct fcache_extent *last, *fe;
	struct list_head *node;
	int ret = 0, i;

	i = 0;
	node = &fe_fs_list;
	last = NULL;
	last_li = NULL;
	while ((node = node->next) != &fe_fs_list) {
		li = list_entry(node, struct fcache_li, fs_list);
		fe = &li->fe;

		if (last) {
			if (last->fs_sector + (last->fs_size >> 9) > fe->fs_sector) {
				ret = 1;
				break;
			}
			if (mergeable(last_li, li))
				(*merges)++;
		}
		last = fe;
		last_li = li;
		i++;
	}

	if (ret)
		goto err;

	i = 0;
	last = NULL;
	last_li = NULL;
	node = &fe_cache_list;
	while ((node = node->next) != &fe_cache_list) {
		li = list_entry(node, struct fcache_li, cache_list);
		fe = &li->fe;

		if (last) {
			if (last->cache_sector + (last->fs_size >> 9) > fe->cache_sector) {
				ret = 1;
				break;
			}
			free(last_li);
		}
		last = fe;
		last_li = li;
		i++;
	}

err:
	if (ret) {
		printf("Overlap: %llu/%u -> %llu/%u\n", (unsigned long long) last->fs_sector, last->fs_size, (unsigned long long) fe->fs_sector, fe->fs_size);
		return 1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	union {
		struct fcache_header hdr;
		char page[PAGE_SIZE];
	} *a;
	struct fcache_header *hdr;
	unsigned int extents;
	int merges = 0;

	if (argc < 2) {
		printf("%s: <dev> [<fsdev]\n", argv[0]);
		return 1;
	}

	fd = open(argv[1], O_RDONLY | O_LARGEFILE | O_DIRECT);
	if (fd < 0)
		return error("open");

	if (argc > 2) {
		fsfd = open(argv[2], O_RDONLY | O_LARGEFILE | O_DIRECT);
		if (fsfd < 0)
			return error("open");
	}

	a = ALIGN(malloc(2 * PAGE_SIZE - 1));
	hdr = &a->hdr;

	if (read(fd, &a->hdr, sizeof(*a)) != sizeof(*a))
		return error("read hdr");

	if (hdr->magic != FCACHE_MAGIC) {
		printf("bad magic %x\n", hdr->magic);
		return 1;
	}
	if (hdr->version != FCACHE_VERSION) {
		printf("bad version %x\n", hdr->version);
		return 1;
	}

	printf("File system: %s (%llu/%llu)\n", hdr->fs_dev, (unsigned long long) hdr->fs_start_sector, (unsigned long long) hdr->fs_sectors);
	printf("\tExtents:\t%u, %u in use (%3.2f%% full)\n", hdr->max_extents, hdr->nr_extents, (float) 100 * hdr->nr_extents / (float) hdr->max_extents);
	printf("\tBlocks:\t\t%llu, %llu in use (%3.2f%% full)\n", (unsigned long long) hdr->cache_blocks, (unsigned long long) hdr->cache_blocks_used, (float) 100 * hdr->cache_blocks_used / (float) hdr->cache_blocks);
	printf("\tExtent offset:\t%u blocks\n", hdr->extent_offset);
	extents = hdr->nr_extents;
	show_extents(hdr);
	printf("1. Found %d good, %d bad extents of %d:", good_extents, bad_extents, extents);
	if (!bad_extents)
		printf("\tOK\n");
	else
		printf("\tFAIL\n");
	if (!verify_layout(&merges)) {
		printf("2. File system layout verified:\t\t\tOK (%d mergeable)\n", merges);
		printf("3. Cache layout verified:\t\t\tOK\n");
	}

	close(fd);
	if (fsfd)
		close(fsfd);
	return 0;
}
